
/*******************************************************************/
/*                                                                 */
/*     File: reorder.C                                             */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Thu Jun 18 15:51:00 2009                              */
/* Modified: Fri Jul 17 17:51:34 2009 (schmid)                     */
/* Copyright: Institut fuer maschinelle Sprachverarbeitung         */
/*            Universitaet Stuttgart                               */
/*                                                                 */
/*******************************************************************/


#include "tree.h"


/*******************************************************************/
/*                                                                 */
/*  Tree::move_node                                                */
/*                                                                 */
/*******************************************************************/

void Tree::move_node( int from, int to )

{
  Tree *child = daughters[from];

  if (from < to)
    for( int i=from; i<to; i++)
      daughters[i] = daughters[i+1];
  else
    for( int i=from; i>to; i--)
      daughters[i] = daughters[i-1];

  daughters[to] = child;
}


/*******************************************************************/
/*                                                                 */
/*  Tree::is_mod                                                   */
/*                                                                 */
/*******************************************************************/

bool Tree::is_mod( bool pp_also)

{
  return ((matching_base_cat("PP-MO") && pp_also) ||
	  label[0] == '$' || 
	  matching_base_cat("ADV-MO") ||
	  matching_base_cat("ADJD-MO") ||
	  matching_base_cat("AVP-MO") ||
	  matching_base_cat("PTKNEG") ||
	  matching_base_cat("PWAV") ||
	  matching_base_cat("PWS") ||
	  matching_base_cat("NP-MO") ||
	  matching_base_cat("VP-MO") ||
	  matching_base_cat("CARD-MO") ||
	  matching_base_cat("PRELS") ||
	  matching_base_cat("PROAV") ||
	  matching_base_cat("KOUI") ||
	  matching_base_cat("KOUS") ||
	  matching_base_cat("KOKOM") ||
	  matching_base_cat("KON") ||
	  (matching_base_cat("NP") &&
	   (has_feature("rel") || has_feature("wh"))));
}


/*******************************************************************/
/*                                                                 */
/*  Tree::has_feature                                              */
/*                                                                 */
/*******************************************************************/

bool Tree::has_feature( const char *f )

{
  for( char *p=label; *p; p++ )
    if (*p == '/' && strncmp(p+1,f,strlen(f)) == 0 &&
	(*(p+4) == 0 || isalnum(*(p+4))))
      return true;
  return false;
}


/*******************************************************************/
/*                                                                 */
/*  Tree::reverse_order                                            */
/*                                                                 */
/*******************************************************************/

void Tree::reverse_order( int i, int k )

{
  for( ; i<k; i++, k-- ) {
    Tree *t = daughters[i];
    daughters[i] = daughters[k];
    daughters[k] = t;
  }
}


/*******************************************************************/
/*                                                                 */
/*  Tree::reorder                                                  */
/*                                                                 */
/*******************************************************************/

void Tree::reorder()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->reorder();

  if (matching_base_cat("S") || matching_base_cat("VP")) {

    // Move adjunct clauses to the end of the sentence
    int k=daughters.size()-1;
    for( int i=0; i<k; i++ )
      if (daughters[i]->matching_base_cat("S-MO") ||
	  daughters[i]->matching_base_cat("VP-MO")
// 	  ||
// 	  (daughters[i]->matching_base_cat("PP-MO") &&
// 	   matching_category("S-OC"))
	  )
	{
	  move_node(i, k--);
	  if (strcmp(daughters[i]->label,"$,")==0 && i < k)
	    move_node(i, k--);
	  i--;
	}

    size_t i=0;
    // find the subject and move it in front
    if (matching_base_cat("S")) {

      // skip over modifiers
      while (i<daughters.size() && daughters[i]->is_mod(true))
	i++;

      for( size_t k=i; k<daughters.size(); k++ )
	if (daughters[k]->matching_base_cat("NP-SB") ||
	    daughters[k]->matching_base_cat("NP-EP") ||
	    daughters[k]->matching_base_cat("CNP-SB") ||
	    daughters[k]->matching_base_cat("PN-SB") ||
	    daughters[k]->matching_base_cat("VP-SB") ||
	    daughters[k]->matching_base_cat("ADJA-SB"))
	  {
	    move_node(k,i++);
	  }
    }
    
    while (i<daughters.size() && daughters[i]->is_mod(false))
      i++;

    // find the main verb and move it in front
    for( size_t k=i; k<daughters.size(); k++ )
      if (daughters[k]->matching_base_cat("V?FIN") ||
	  daughters[k]->matching_base_cat("VZ-HD") ||
	  daughters[k]->matching_base_cat("VVIZU-HD"))
	{
	  move_node(k,i++);
	  break;
	}
    
    // skip over modifiers
    while (i<daughters.size() && daughters[i]->is_mod(false))
      i++;

    // Move adverbs and reflexives in front
    for( size_t k=i; k<daughters.size(); k++ )
      if (daughters[k]->matching_base_cat("PRF") ||
	  daughters[k]->matching_base_cat("PTKNEG") ||
	  daughters[k]->matching_base_cat("ADV-MO") ||
	  daughters[k]->matching_base_cat("ADJD-MO") ||
	  daughters[k]->matching_base_cat("AVP-MO") ||
	  (daughters[k]->matching_base_cat("NP") &&
	   daughters[k]->daughters.size() == 1 &&
	   (daughters[k]->daughters[0]->matching_base_cat("PPER") ||
	    daughters[k]->daughters[0]->matching_base_cat("PDS"))))
	{
	  move_node(k, i++);
	}
    
    // find an optional non-finite verb (phrase) and move it in front
    
    for( size_t k=i; k<daughters.size(); k++ )
      if (daughters[k]->matching_base_cat("V?PP") ||
	  daughters[k]->matching_base_cat("V?INF") ||
	  daughters[k]->matching_base_cat("PTKVZ"))
	{
	  move_node(k,i++);
	  break;
	}
    
    for( size_t k=i; k<daughters.size(); k++ )
      if (daughters[k]->matching_category("VP-OC") ||
	  daughters[k]->matching_category("VP-PD"))
	{
	  move_node(k,i++);
	  break;
	}
    
    size_t l;
    for( l=daughters.size()-1; l>i; l-- ) {
      vector<char*> &s=split_name(daughters[l]->label, 3);
      if (//!strncmp(s[1], "-MO", 3) == 0 &&
	  (strncmp(s[0], "NP", 2) == 0 ||
	   strncmp(s[0], "PP", 2) == 0 ||
	   strncmp(s[0], "PRF", 3) == 0 ||
	   strncmp(s[0], "PPER", 4) == 0))
	break;
    }
    
    reverse_order( i, l);
  }

  // move complex adjectival phrases to the end of NPs
  if (matching_base_cat("PP") || matching_base_cat("NP"))
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("AP") ||
	  daughters[i]->matching_base_cat("VP")) {
	Tree *d=daughters[i];
	for( size_t k=0; k<d->daughters.size(); k++ )
	  if ((d->daughters[k]->matching_base_cat("NP") ||
	       d->daughters[k]->matching_base_cat("PP")) &&
	      !d->daughters[k]->is_trace() &&
	      !d->daughters[k]->matching_base_cat("NP-AMS"))
	    {
	      if (d->matching_base_cat("AP")) {
		for( ; k<d->daughters.size(); k++ )
		  if (d->daughters[k]->matching_base_cat("ADJA") ||
		      d->daughters[k]->matching_base_cat("VVPP"))
		    {
		      d->move_node( k, 0 );
		      break;
		    }
		      
	      }
	      move_node( i, daughters.size()-1);
	      break;
	    }
	break;
      }
}


/*******************************************************************/
/*                                                                 */
/*  Tree::is_wh_phrase                                             */
/*                                                                 */
/*******************************************************************/

bool Tree::is_wh_phrase()

{
  return (matching_base_cat("PWAV") ||
	  matching_base_cat("PWS") ||
	  has_feature("rel") ||
	  has_feature("wh"));
}


/*******************************************************************/
/*                                                                 */
/*  Tree::find_filler                                              */
/*                                                                 */
/*******************************************************************/

Tree *Tree::find_filler( char *index )

{
  Tree *result = NULL;
  for( size_t i=0; !result && i<daughters.size(); i++ ) {
    // Is the current node a filler?
    char *s;
    Tree *d=daughters[i];
    if (d->daughters.size() > 0 && (s = find_index(d->label)) &&
	!d->is_wh_phrase() && !d->matching_base_cat("AP-CC")) {
      *(s++) = 0; // delete the index
      strcpy(index, s);
      Tree *result = d;
      
      return result;
    }
    
    result = d->find_filler( index );
  }
  return result;
}


/*******************************************************************/
/*                                                                 */
/*  Tree::fill_gap                                                 */
/*                                                                 */
/*******************************************************************/

bool Tree::fill_gap( Tree *node, char *index )

{
  for( size_t i=0; i<daughters.size(); i++ ) {
    Tree *d=daughters[i];
    if (d->matching_category("*T*") && strcmp(d->label+4, index) == 0 &&
	!(node->matching_category("VP-RE") &&
	  mother->matching_category("NP-SB")))
      {
	// remove the filler node
	Tree *m=node->mother;
	for( size_t i=0; i<m->daughters.size(); i++ )
	  if (m->daughters[i] == node) {
	    for( i++ ; i<m->daughters.size(); i++ )
	      m->daughters[i-1] = m->daughters[i];
	    m->daughters.resize(m->daughters.size()-1);
	  }
	
	// replace the trace (VP-OC *T*-3)
	for( size_t i=0; i<mother->daughters.size(); i++ )
	  if (mother->daughters[i] == this) {
	    mother->daughters[i] = node;
	    delete this;
	    return true;
	  }
      }
      
    if (d->fill_gap( node, index ))
      return true;
  }
  return false;
}


/*******************************************************************/
/*                                                                 */
/*  Tree::fill_gaps                                                */
/*                                                                 */
/*******************************************************************/

void Tree::fill_gaps()

{
  Tree *filler;
  char index[1000];
  while ((filler = find_filler( index )))
    if (!filler->matching_category("S-CC"))
      fill_gap( filler, index );
}
